@@ -1,4 +1,4 @@
-# Copyright 2020 Canonical Ltd.
+# Copyright 2020-2021 Canonical Ltd.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,6 +12,8 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+"""The Operator Framework infrastructure."""
+
 import collections
 import collections.abc
 import inspect
@@ -24,13 +26,11 @@ import pdb
 import re
 import sys
 import types
+import typing
 import weakref
 
 from ops import charm
-from ops.storage import (
-    NoSnapshotError,
-    SQLiteStorage,
-)
+from ops.storage import NoSnapshotError, SQLiteStorage
 
 logger = logging.getLogger(__name__)
 
@@ -67,6 +67,7 @@ class Handle:
                 self._path = "{}".format(kind)
 
     def nest(self, kind, key):
+        """Create a new handle as child of the current one."""
         return Handle(self, kind, key)
 
     def __hash__(self):
@@ -80,22 +81,27 @@ class Handle:
 
     @property
     def parent(self):
+        """Return own parent handle."""
         return self._parent
 
     @property
     def kind(self):
+        """Return the handle's kind."""
         return self._kind
 
     @property
     def key(self):
+        """Return the handle's key."""
         return self._key
 
     @property
     def path(self):
+        """Return the handle's path."""
         return self._path
 
     @classmethod
     def from_path(cls, path):
+        """Build a handle from the indicated path."""
         handle = None
         for pair in path.split("/"):
             pair = pair.split("[")
@@ -115,6 +121,10 @@ class Handle:
 
 
 class EventBase:
+    """The base for all the different Events.
+
+    Inherit this and override 'snapshot' and 'restore' methods to build a custom event.
+    """
 
     def __init__(self, handle):
         self.handle = handle
@@ -124,6 +134,55 @@ class EventBase:
         return "<%s via %s>" % (self.__class__.__name__, self.handle)
 
     def defer(self):
+        """Defer the event to the future.
+
+        Deferring an event from a handler puts that handler into a queue, to be
+        called again the next time the charm is invoked. This invocation may be
+        the result of an action, or any event other than metric events. The
+        queue of events will be dispatched before the new event is processed.
+
+        From the above you may deduce, but it's important to point out:
+
+        * ``defer()`` does not interrupt the execution of the current event
+          handler. In almost all cases, a call to ``defer()`` should be followed
+          by an explicit ``return`` from the handler;
+
+        * the re-execution of the deferred event handler starts from the top of
+          the handler method (not where defer was called);
+
+        * only the handlers that actually called ``defer()`` are called again
+          (that is: despite talking about “deferring an event” it is actually
+          the handler/event combination that is deferred); and
+
+        * any deferred events get processed before the event (or action) that
+          caused the current invocation of the charm.
+
+        The general desire to call ``defer()`` happens when some precondition
+        isn't yet met. However, care should be exercised as to whether it is
+        better to defer this event so that you see it again, or whether it is
+        better to just wait for the event that indicates the precondition has
+        been met.
+
+        For example, if ``config-changed`` is fired, and you are waiting for
+        different config, there is no reason to defer the event because there
+        will be a *different* ``config-changed`` event when the config actually
+        changes, rather than checking to see if maybe config has changed prior
+        to every other event that occurs.
+
+        Similarly, if you need 2 events to occur before you are ready to
+        proceed (say event A and B). When you see event A, you could chose to
+        ``defer()`` it because you haven't seen B yet. However, that leads to:
+
+        1. event A fires, calls defer()
+
+        2. event B fires, event A handler is called first, still hasn't seen B
+           happen, so is deferred again. Then B happens, which progresses since
+           it has seen A.
+
+        3. At some future time, event C happens, which also checks if A can
+           proceed.
+
+        """
         logger.debug("Deferring %s.", self)
         self.deferred = True
 
@@ -191,6 +250,7 @@ class EventSource:
 
 
 class BoundEvent:
+    """Event bound to an Object."""
 
     def __repr__(self):
         return '<BoundEvent {} bound to {}.{} at {}>'.format(
@@ -250,8 +310,8 @@ class _Metaclass(type):
 
     """
 
-    def __new__(typ, *a, **kw):
-        k = super().__new__(typ, *a, **kw)
+    def __new__(cls, *a, **kw):
+        k = super().__new__(cls, *a, **kw)
         # k is now the Object-derived class; loop over its class attributes
         for n, v in vars(k).items():
             # we could do duck typing here if we want to support
@@ -263,7 +323,24 @@ class _Metaclass(type):
 
 
 class Object(metaclass=_Metaclass):
+    """Initialize an Object as a new leaf in :class:`Framework`, identified by `key`.
 
+    Args:
+        parent: parent node in the tree.
+        key: unique identifier for this object.
+
+    Every object belongs to exactly one framework.
+
+    Every object has a parent, which might be a framework.
+
+    We track a "path to object," which is the path to the parent, plus the object's unique
+    identifier. Event handlers use this identity to track the destination of their events, and the
+    Framework uses this id to track persisted state between event executions.
+
+    The Framework should raise an error if it ever detects that two objects with the same id have
+    been created.
+
+    """
     handle_kind = HandleKind()
 
     def __init__(self, parent, key):
@@ -283,6 +360,7 @@ class Object(metaclass=_Metaclass):
 
     @property
     def model(self):
+        """Shortcut for more simple access the model."""
         return self.framework.model
 
 
@@ -336,26 +414,32 @@ class ObjectEvents(Object):
         event_descriptor._set_name(cls, event_kind)
         setattr(cls, event_kind, event_descriptor)
 
-    def events(self):
-        """Return a mapping of event_kinds to bound_events for all available events.
-        """
-        events_map = {}
+    def _event_kinds(self):
+        event_kinds = []
         # We have to iterate over the class rather than instance to allow for properties which
         # might call this method (e.g., event views), leading to infinite recursion.
         for attr_name, attr_value in inspect.getmembers(type(self)):
             if isinstance(attr_value, EventSource):
                 # We actually care about the bound_event, however, since it
                 # provides the most info for users of this method.
-                event_kind = attr_name
-                bound_event = getattr(self, event_kind)
-                events_map[event_kind] = bound_event
-        return events_map
+                event_kinds.append(attr_name)
+        return event_kinds
+
+    def events(self):
+        """Return a mapping of event_kinds to bound_events for all available events."""
+        return {event_kind: getattr(self, event_kind) for event_kind in self._event_kinds()}
 
     def __getitem__(self, key):
         return PrefixedEvents(self, key)
 
+    def __repr__(self):
+        k = type(self)
+        event_kinds = ', '.join(sorted(self._event_kinds()))
+        return '<{}.{}: {}>'.format(k.__module__, k.__qualname__, event_kinds)
+
 
 class PrefixedEvents:
+    """Events to be found in all events using a specific prefix."""
 
     def __init__(self, emitter, key):
         self._emitter = emitter
@@ -366,19 +450,21 @@ class PrefixedEvents:
 
 
 class PreCommitEvent(EventBase):
-    pass
+    """Events that will be emitted first on commit."""
 
 
 class CommitEvent(EventBase):
-    pass
+    """Events that will be emitted second on commit."""
 
 
 class FrameworkEvents(ObjectEvents):
+    """Manager of all framework events."""
     pre_commit = EventSource(PreCommitEvent)
     commit = EventSource(CommitEvent)
 
 
 class NoTypeError(Exception):
+    """No class to hold it was found when restoring an event."""
 
     def __init__(self, handle_path):
         self.handle_path = handle_path
@@ -401,6 +487,7 @@ _event_regex = r'^(|.*/)on/[a-zA-Z_]+\[\d+\]$'
 
 
 class Framework(Object):
+    """Main interface to from the Charm to the Operator Framework internals."""
 
     on = FrameworkEvents()
 
@@ -437,18 +524,36 @@ class Framework(Object):
             self._stored = StoredStateData(self, '_stored')
             self._stored['event_count'] = 0
 
-        # Hook into builtin breakpoint, so if Python >= 3.7, devs will be able to just do
-        # breakpoint(); if Python < 3.7, this doesn't affect anything
-        sys.breakpointhook = self.breakpoint
-
         # Flag to indicate that we already presented the welcome message in a debugger breakpoint
         self._breakpoint_welcomed = False
 
-        # Parse once the env var, which may be used multiple times later
+        # Parse the env var once, which may be used multiple times later
         debug_at = os.environ.get('JUJU_DEBUG_AT')
-        self._juju_debug_at = debug_at.split(',') if debug_at else ()
+        self._juju_debug_at = set(x.strip() for x in debug_at.split(',')) if debug_at else set()
+
+    def set_breakpointhook(self):
+        """Hook into sys.breakpointhook so the builtin breakpoint() works as expected.
+
+        This method is called by ``main``, and is not intended to be
+        called by users of the framework itself outside of perhaps
+        some testing scenarios.
+
+        It returns the old value of sys.excepthook.
+
+        The breakpoint function is a Python >= 3.7 feature.
+
+        This method was added in ops 1.0; before that, it was done as
+        part of the Framework's __init__.
+        """
+        old_breakpointhook = getattr(sys, 'breakpointhook', None)
+        if old_breakpointhook is not None:
+            # Hook into builtin breakpoint, so if Python >= 3.7, devs will be able to just do
+            # breakpoint()
+            sys.breakpointhook = self.breakpoint
+        return old_breakpointhook
 
     def close(self):
+        """Close the underlying backends."""
         self._storage.close()
 
     def _track(self, obj):
@@ -466,6 +571,7 @@ class Framework(Object):
         self._objects.pop(obj.handle.path, None)
 
     def commit(self):
+        """Save changes to the underlying backends."""
         # Give a chance for objects to persist data they want to before a commit is made.
         self.on.pre_commit.emit()
         # Make sure snapshots are saved by instances of StoredStateData. Any possible state
@@ -476,6 +582,7 @@ class Framework(Object):
         self._storage.commit()
 
     def register_type(self, cls, parent, kind=None):
+        """Register a type to a handle."""
         if parent and not isinstance(parent, Handle):
             parent = parent.handle
         if parent:
@@ -515,6 +622,7 @@ class Framework(Object):
         self._storage.save_snapshot(value.handle.path, data)
 
     def load_snapshot(self, handle):
+        """Load a persistent snapshot."""
         parent_path = None
         if handle.parent:
             parent_path = handle.parent.path
@@ -530,18 +638,19 @@ class Framework(Object):
         return obj
 
     def drop_snapshot(self, handle):
+        """Discard a persistent snapshot."""
         self._storage.drop_snapshot(handle.path)
 
-    def observe(self, bound_event: BoundEvent, observer: types.MethodType):
+    def observe(self, bound_event: BoundEvent, observer: typing.Callable[[typing.Any], None]):
         """Register observer to be called when bound_event is emitted.
 
         The bound_event is generally provided as an attribute of the object that emits
-        the event, and is created in this style:
+        the event, and is created in this style::
 
             class SomeObject:
                 something_happened = Event(SomethingHappened)
 
-        That event may be observed as:
+        That event may be observed as::
 
             framework.observe(someobj.something_happened, self._on_something_happened)
 
@@ -605,7 +714,6 @@ class Framework(Object):
 
     def _emit(self, event):
         """See BoundEvent.emit for the public way to call this."""
-
         saved = False
         event_path = event.handle.path
         event_kind = event.handle.kind
@@ -664,7 +772,9 @@ class Framework(Object):
                 if custom_handler:
                     event_is_from_juju = isinstance(event, charm.HookEvent)
                     event_is_action = isinstance(event, charm.ActionEvent)
-                    if (event_is_from_juju or event_is_action) and 'hook' in self._juju_debug_at:
+                    if (
+                        event_is_from_juju or event_is_action
+                    ) and self._juju_debug_at.intersection({'all', 'hook'}):
                         # Present the welcome message and run under PDB.
                         self._show_debug_code_message()
                         pdb.runcall(custom_handler, event)
@@ -746,6 +856,7 @@ class Framework(Object):
 
 
 class StoredStateData(Object):
+    """Manager of the stored data."""
 
     def __init__(self, parent, attr_name):
         super().__init__(parent, attr_name)
@@ -763,19 +874,23 @@ class StoredStateData(Object):
         return key in self._cache
 
     def snapshot(self):
+        """Return the current state."""
         return self._cache
 
     def restore(self, snapshot):
+        """Restore current state to the given snapshot."""
         self._cache = snapshot
         self.dirty = False
 
     def on_commit(self, event):
+        """Save changes to the storage backend."""
         if self.dirty:
             self.framework.save_snapshot(self)
             self.dirty = False
 
 
 class BoundStoredState:
+    """Stored state data bound to a specific Object."""
 
     def __init__(self, parent, attr_name):
         parent.framework.register_type(StoredStateData, parent)
@@ -814,7 +929,7 @@ class BoundStoredState:
         self._data[key] = _unwrap_stored(self._data, value)
 
     def set_default(self, **kwargs):
-        """"Set the value of any given key if it has not already been set"""
+        """Set the value of any given key if it has not already been set."""
         for k, v in kwargs.items():
             if k not in self._data:
                 self._data[k] = v
@@ -913,7 +1028,16 @@ def _unwrap_stored(parent_data, value):
     return value
 
 
+def _wrapped_repr(obj):
+    t = type(obj)
+    if obj._under:
+        return "{}.{}({!r})".format(t.__module__, t.__name__, obj._under)
+    else:
+        return "{}.{}()".format(t.__module__, t.__name__)
+
+
 class StoredDict(collections.abc.MutableMapping):
+    """A dict-like object that uses the StoredState as backend."""
 
     def __init__(self, stored_data, under):
         self._stored_data = stored_data
@@ -944,8 +1068,11 @@ class StoredDict(collections.abc.MutableMapping):
         else:
             return NotImplemented
 
+    __repr__ = _wrapped_repr
+
 
 class StoredList(collections.abc.MutableSequence):
+    """A list-like object that uses the StoredState as backend."""
 
     def __init__(self, stored_data, under):
         self._stored_data = stored_data
@@ -966,10 +1093,12 @@ class StoredList(collections.abc.MutableSequence):
         return len(self._under)
 
     def insert(self, index, value):
+        """Insert value before index."""
         self._under.insert(index, value)
         self._stored_data.dirty = True
 
     def append(self, value):
+        """Append value to the end of the list."""
         self._under.append(value)
         self._stored_data.dirty = True
 
@@ -1013,18 +1142,29 @@ class StoredList(collections.abc.MutableSequence):
         else:
             return NotImplemented
 
+    __repr__ = _wrapped_repr
+
 
 class StoredSet(collections.abc.MutableSet):
+    """A set-like object that uses the StoredState as backend."""
 
     def __init__(self, stored_data, under):
         self._stored_data = stored_data
         self._under = under
 
     def add(self, key):
+        """Add a key to a set.
+
+        This has no effect if the key is already present.
+        """
         self._under.add(key)
         self._stored_data.dirty = True
 
     def discard(self, key):
+        """Remove a key from a set if it is a member.
+
+        If the key is not a member, do nothing.
+        """
         self._under.discard(key)
         self._stored_data.dirty = True
 
@@ -1071,3 +1211,5 @@ class StoredSet(collections.abc.MutableSet):
             return self._under == other
         else:
             return NotImplemented
+
+    __repr__ = _wrapped_repr
