@@ -1,4 +1,4 @@
-# Copyright 2019 Canonical Ltd.
+# Copyright 2019-2021 Canonical Ltd.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,9 +15,10 @@
 """Representations of Juju's model, application, unit, and other entities."""
 
 import datetime
-import decimal
 import ipaddress
 import json
+import logging
+import math
 import os
 import re
 import shutil
@@ -25,16 +26,19 @@ import tempfile
 import time
 import typing
 import weakref
-
 from abc import ABC, abstractmethod
 from collections.abc import Mapping, MutableMapping
 from pathlib import Path
-from subprocess import run, PIPE, CalledProcessError
+from subprocess import PIPE, CalledProcessError, run
 
-from ops._private import yaml
-from ops.jujuversion import JujuVersion
 import ops
 import ops.pebble as pebble
+from ops._private import yaml
+from ops.jujuversion import JujuVersion
+
+logger = logging.getLogger(__name__)
+
+MAX_LOG_LINE_LEN = 131071  # Max length of strings to pass to subshell.
 
 
 class Model:
@@ -106,6 +110,14 @@ class Model:
         """
         return self._backend.model_name
 
+    @property
+    def uuid(self) -> str:
+        """Return the identifier of the Model that this unit is running in.
+
+        This is read from the environment variable ``JUJU_MODEL_UUID``.
+        """
+        return self._backend.model_uuid
+
     def get_unit(self, unit_name: str) -> 'Unit':
         """Get an arbitrary unit by name.
 
@@ -242,6 +254,24 @@ class Application:
         self._backend.status_set(value.name, value.message, is_app=True)
         self._status = value
 
+    def planned_units(self) -> int:
+        """Get the number of units that Juju has "planned" for this application.
+
+        E.g., if an operator ran "juju deploy foo", then "juju add-unit -n 2 foo", the
+        planned unit count for foo would be 3.
+
+        We deliberately do not attempt to inspect whether these units are actually running
+        or not at present.
+
+        This only works for this charm's application -- the unit agent isn't able to get
+        planned units for other applications in the model.
+        """
+        if not self._is_our_app:
+            raise RuntimeError(
+                'cannot get planned units for a remote application {}.'.format(self))
+
+        return self._backend.planned_units()
+
     def __repr__(self):
         return '<{}.{} {}>'.format(type(self).__module__, type(self).__name__, self.name)
 
@@ -268,7 +298,7 @@ class Unit:
         self._is_our_unit = self.name == self._backend.unit_name
         self._status = None
 
-        if self._is_our_unit:
+        if self._is_our_unit and hasattr(meta, "containers"):
             self._containers = ContainerMapping(meta.containers, backend)
 
     def _invalidate(self):
@@ -634,7 +664,8 @@ class Relation:
     """Represents an established relation between this application and another application.
 
     This class should not be instantiated directly, instead use :meth:`Model.get_relation`
-    or :attr:`ops.charm.RelationEvent.relation`.
+    or :attr:`ops.charm.RelationEvent.relation`. This is principally used by
+    :class:`ops.charm.RelationMeta` to represent the relationships between charms.
 
     Attributes:
         name: The name of the local endpoint of the relation (eg 'db')
@@ -920,10 +951,10 @@ class Resources:
         """Fetch the resource from the controller or store.
 
         If successfully fetched, this returns a Path object to where the resource is stored
-        on disk, otherwise it raises a ModelError.
+        on disk, otherwise it raises a NameError.
         """
         if name not in self._paths:
-            raise RuntimeError('invalid resource name: {}'.format(name))
+            raise NameError('invalid resource name: {}'.format(name))
         if self._paths[name] is None:
             self._paths[name] = Path(self._backend.resource_get(name))
         return self._paths[name]
@@ -990,6 +1021,13 @@ class StorageMapping(Mapping):
                               ' it is not present in the charm metadata').format(storage_name))
         self._backend.storage_add(storage_name, count)
 
+    def _invalidate(self, storage_name):
+        """Remove an entry from the storage map.
+
+        Not meant to be used by charm authors -- this exists mainly for testing purposes.
+        """
+        self._storage_map[storage_name] = None
+
 
 class Storage:
     """Represents a storage as defined in metadata.yaml.
@@ -1006,13 +1044,22 @@ class Storage:
         self._location = None
 
     @property
-    def location(self):
+    def location(self) -> Path:
         """Return the location of the storage."""
         if self._location is None:
             raw = self._backend.storage_get('{}/{}'.format(self.name, self.id), "location")
             self._location = Path(raw)
         return self._location
 
+    @location.setter
+    def location(self, location: str) -> None:
+        """Sets the location for use in events.
+
+        For :class:`StorageAttachedEvent` and :class:`StorageDetachingEvent` in case
+        the actual details are gone from Juju by the time of a dynamic lookup.
+        """
+        self._location = Path(location)
+
 
 class Container:
     """Represents a named container in a unit.
@@ -1034,23 +1081,94 @@ class Container:
 
     @property
     def pebble(self) -> 'pebble.Client':
-        """Return the low-level Pebble client instance for this container."""
+        """The low-level :class:`ops.pebble.Client` instance for this container."""
         return self._pebble
 
+    def can_connect(self) -> bool:
+        """Report whether the Pebble API is reachable in the container.
+
+        :meth:`can_connect` returns a bool that indicates whether the Pebble API is available at
+        the time the method is called. It does not guard against the Pebble API becoming
+        unavailable, and should be treated as a 'point in time' status only.
+
+        If the Pebble API later fails, serious consideration should be given as to the reason for
+        this.
+
+        Example::
+
+            container = self.unit.get_container("example")
+            if container.can_connect():
+                try:
+                    c.pull('/does/not/exist')
+                except ProtocolError, PathError:
+                    # handle it
+            else:
+                event.defer()
+        """
+        try:
+            # TODO: This call to `get_system_info` should be replaced with a call to a more
+            # appropriate endpoint that has stronger connotations of what constitutes a Pebble
+            # instance that is in fact 'ready'.
+            self._pebble.get_system_info()
+        except pebble.ConnectionError as e:
+            logger.debug("Pebble API is not ready; ConnectionError: %s", e.message())
+            return False
+        except FileNotFoundError as e:
+            # In some cases, charm authors can attempt to hit the Pebble API before it has had the
+            # chance to create the UNIX socket in the shared volume.
+            logger.debug("Pebble API is not ready; UNIX socket not found:", str(e))
+            return False
+        except pebble.APIError as e:
+            # An API error is only raised when the Pebble API returns invalid JSON, or the response
+            # cannot be read. Both of these are a likely indicator that something is wrong.
+            logger.warning("Pebble API is not ready; APIError: %s", str(e))
+            return False
+        return True
+
     def autostart(self):
         """Autostart all services marked as startup: enabled."""
         self._pebble.autostart_services()
 
+    def replan(self):
+        """Replan all services: restart changed services and start startup-enabled services."""
+        self._pebble.replan_services()
+
     def start(self, *service_names: str):
         """Start given service(s) by name."""
+        if not service_names:
+            raise TypeError('start expected at least 1 argument, got 0')
+
         self._pebble.start_services(service_names)
 
+    def restart(self, *service_names: str):
+        """Restart the given service(s) by name."""
+        if not service_names:
+            raise TypeError('restart expected at least 1 argument, got 0')
+
+        try:
+            self._pebble.restart_services(service_names)
+        except pebble.APIError as e:
+            if e.code != 400:
+                raise e
+            # support old Pebble instances that don't support the "restart" action
+            stop = tuple(s.name for s in self.get_services(*service_names).values()
+                         if s.is_running())
+            if stop:
+                self._pebble.stop_services(stop)
+
+            self._pebble.start_services(service_names)
+
     def stop(self, *service_names: str):
         """Stop given service(s) by name."""
+        if not service_names:
+            raise TypeError('stop expected at least 1 argument, got 0')
+
         self._pebble.stop_services(service_names)
 
-    def add_layer(self, label: str, layer: typing.Union[str, typing.Dict, 'pebble.Layer'], *,
-                  combine: bool = False):
+    # TODO(benhoyt) - should be: layer: typing.Union[str, typing.Dict, 'pebble.Layer'],
+    # but this breaks on Python 3.5.2 (the default on Xenial). See:
+    # https://github.com/canonical/operator/issues/517
+    def add_layer(self, label: str, layer, *, combine: bool = False):
         """Dynamically add a new layer onto the Pebble configuration layers.
 
         Args:
@@ -1123,10 +1241,12 @@ class Container:
             make_dirs: If True, create parent directories if they don't exist.
             permissions: Permissions (mode) to create file with (Pebble default
                 is 0o644).
-            user_id: UID for file.
-            user: Username for file (user_id takes precedence).
-            group_id: GID for file.
-            group: Group name for file (group_id takes precedence).
+            user_id: User ID (UID) for file.
+            user: Username for file. User's UID must match user_id if both are
+                specified.
+            group_id: Group ID (GID) for file.
+            group: Group name for file. Group's GID must match group_id if
+                both are specified.
         """
         self._pebble.push(path, source, encoding=encoding, make_dirs=make_dirs,
                           permissions=permissions, user_id=user_id, user=user,
@@ -1134,13 +1254,16 @@ class Container:
 
     def list_files(self, path: str, *, pattern: str = None,
                    itself: bool = False) -> typing.List['pebble.FileInfo']:
-        """Return list of file information from given path on remote system.
+        """Return list of directory entries from given path on remote system.
+
+        Despite the name, this method returns a list of files *and*
+        directories, similar to :func:`os.listdir` or :func:`os.scandir`.
 
         Args:
             path: Path of the directory to list, or path of the file to return
                 information about.
             pattern: If specified, filter the list to just the files that match,
-                for example "*.txt".
+                for example ``*.txt``.
             itself: If path refers to a directory, return information about the
                 directory itself, rather than its contents.
         """
@@ -1156,10 +1279,12 @@ class Container:
             make_parents: If True, create parent directories if they don't exist.
             permissions: Permissions (mode) to create directory with (Pebble
                 default is 0o755).
-            user_id: UID for directory.
-            user: Username for directory (user_id takes precedence).
-            group_id: GID for directory.
-            group: Group name for directory (group_id takes precedence).
+            user_id: User ID (UID) for directory.
+            user: Username for directory. User's UID must match user_id if
+                both are specified.
+            group_id: Group ID (GID) for directory.
+            group: Group name for directory. Group's GID must match group_id
+                if both are specified.
         """
         self._pebble.make_dir(path, make_parents=make_parents, permissions=permissions,
                               user_id=user_id, user=user, group_id=group_id, group=group)
@@ -1173,6 +1298,61 @@ class Container:
         """
         self._pebble.remove_path(path, recursive=recursive)
 
+    def exec(
+        self,
+        command: typing.List[str],
+        *,
+        environment: typing.Dict[str, str] = None,
+        working_dir: str = None,
+        timeout: float = None,
+        user_id: int = None,
+        user: str = None,
+        group_id: int = None,
+        group: str = None,
+        stdin: typing.Union[str, bytes, typing.TextIO, typing.BinaryIO] = None,
+        stdout: typing.Union[typing.TextIO, typing.BinaryIO] = None,
+        stderr: typing.Union[typing.TextIO, typing.BinaryIO] = None,
+        encoding: str = 'utf-8',
+        combine_stderr: bool = False
+    ) -> 'pebble.ExecProcess':
+        """Execute the given command on the remote system.
+
+        See :meth:`ops.pebble.Client.exec` for documentation of the parameters
+        and return value, as well as examples.
+        """
+        return self._pebble.exec(
+            command,
+            environment=environment,
+            working_dir=working_dir,
+            timeout=timeout,
+            user_id=user_id,
+            user=user,
+            group_id=group_id,
+            group=group,
+            stdin=stdin,
+            stdout=stdout,
+            stderr=stderr,
+            encoding=encoding,
+            combine_stderr=combine_stderr,
+        )
+
+    def send_signal(self, sig: typing.Union[int, str], *service_names: str):
+        """Send the given signal to one or more services.
+
+        Args:
+            sig: Name or number of signal to send, e.g., "SIGHUP", 1, or
+                signal.SIGHUP.
+            service_names: Name(s) of the service(s) to send the signal to.
+
+        Raises:
+            pebble.APIError: If any of the services are not in the plan or are
+                not currently running.
+        """
+        if not service_names:
+            raise TypeError('send_signal expected at least 1 service name, got 0')
+
+        self._pebble.send_signal(sig, service_names)
+
 
 class ContainerMapping(Mapping):
     """Map of container names to Container objects.
@@ -1253,6 +1433,60 @@ class InvalidStatusError(ModelError):
     """Raised if trying to set an Application or Unit status to something invalid."""
 
 
+_ACTION_RESULT_KEY_REGEX = re.compile(r'^[a-z0-9](([a-z0-9-.]+)?[a-z0-9])?$')
+
+
+def _format_action_result_dict(input: dict, parent_key: str = None, output: dict = None) -> dict:
+    """Turn a nested dictionary into a flattened dictionary, using '.' as a key seperator.
+
+    This is used to allow nested dictionaries to be translated into the dotted format required by
+    the Juju `action-set` hook tool in order to set nested data on an action.
+
+    Additionally, this method performs some validation on keys to ensure they only use permitted
+    characters.
+
+    Example::
+
+        >>> test_dict = {'a': {'b': 1, 'c': 2}}
+        >>> _format_action_result_dict(test_dict)
+        {'a.b': 1, 'a.c': 2}
+
+    Arguments:
+        input: The dictionary to flatten
+        parent_key: The string to prepend to dictionary's keys
+        output: The current dictionary to be returned, which may or may not yet be completely flat
+
+    Returns:
+        A flattened dictionary with validated keys
+
+    Raises:
+        ValueError: if the dict is passed with a mix of dotted/non-dotted keys that expand out to
+            result in duplicate keys. For example: {'a': {'b': 1}, 'a.b': 2}. Also raised if a dict
+            is passed with a key that fails to meet the format requirements.
+    """
+    if output is None:
+        output = {}
+
+    for key, value in input.items():
+        # Ensure the key is of a valid format, and raise a ValueError if not
+        if not _ACTION_RESULT_KEY_REGEX.match(key):
+            raise ValueError("key '{!r}' is invalid: must be similar to 'key', 'some-key2', or "
+                             "'some.key'".format(key))
+
+        if parent_key:
+            key = "{}.{}".format(parent_key, key)
+
+        if isinstance(value, MutableMapping):
+            output = _format_action_result_dict(value, key, output)
+        elif key in output:
+            raise ValueError("duplicate key detected in dictionary passed to 'action-set': {!r}"
+                             .format(key))
+        else:
+            output[key] = value
+
+    return output
+
+
 class _ModelBackend:
     """Represents the connection between the Model representation and talking to Juju.
 
@@ -1262,14 +1496,17 @@ class _ModelBackend:
 
     LEASE_RENEWAL_PERIOD = datetime.timedelta(seconds=30)
 
-    def __init__(self, unit_name=None, model_name=None):
+    def __init__(self, unit_name=None, model_name=None, model_uuid=None):
         if unit_name is None:
             self.unit_name = os.environ['JUJU_UNIT_NAME']
         else:
             self.unit_name = unit_name
         if model_name is None:
             model_name = os.environ.get('JUJU_MODEL_NAME')
+        if model_uuid is None:
+            model_uuid = os.environ.get('JUJU_MODEL_UUID')
         self.model_name = model_name
+        self.model_uuid = model_uuid
         self.app_name = self.unit_name.split('/')[0]
 
         self._is_leader = None
@@ -1329,7 +1566,7 @@ class _ModelBackend:
                 return None
             if 'option provided but not defined: --app' in str(e):
                 # "--app" was introduced to relation-list in Juju 2.8.1, so
-                # handle previous verions of Juju gracefully
+                # handle previous versions of Juju gracefully
                 return None
             raise
 
@@ -1465,6 +1702,20 @@ class _ModelBackend:
         return [int(s.split('/')[1]) for s in self._run('storage-list', name,
                                                         return_output=True, use_json=True)]
 
+    def _storage_event_details(self) -> typing.Tuple[int, str]:
+        output = self._run('storage-get', '--help', return_output=True)
+
+        # Match the entire string at once instead of going line by line
+        matcher = re.compile(
+            r'.*^-s\s+\(=\s+(?P<storage_key>.*?)\)\s*?$',
+            re.MULTILINE | re.DOTALL
+        )
+        key = matcher.match(output).groupdict()["storage_key"]
+
+        id = int(key.split("/")[1])
+        location = self.storage_get(key, "location")
+        return id, location
+
     def storage_get(self, storage_name_id, attribute):
         return self._run('storage-get', '-s', storage_name_id, attribute,
                          return_output=True, use_json=True)
@@ -1479,7 +1730,10 @@ class _ModelBackend:
         return self._run('action-get', return_output=True, use_json=True)
 
     def action_set(self, results):
-        self._run('action-set', *["{}={}".format(k, v) for k, v in results.items()])
+        # The Juju action-set hook tool cannot interpret nested dicts, so we use a helper to
+        # flatten out any nested dict structures into a dotted notation, and validate keys.
+        flat_results = _format_action_result_dict(results)
+        self._run('action-set', *["{}={}".format(k, v) for k, v in flat_results.items()])
 
     def action_log(self, message):
         self._run('action-log', message)
@@ -1490,8 +1744,24 @@ class _ModelBackend:
     def application_version_set(self, version):
         self._run('application-version-set', '--', version)
 
+    @classmethod
+    def log_split(cls, message, max_len=MAX_LOG_LINE_LEN):
+        """Helper to handle log messages that are potentially too long.
+
+        This is a generator that splits a message string into multiple chunks if it is too long
+        to safely pass to bash. Will only generate a single entry if the line is not too long.
+        """
+        if len(message) > max_len:
+            yield "Log string greater than {}. Splitting into multiple chunks: ".format(max_len)
+
+        while message:
+            yield message[:max_len]
+            message = message[max_len:]
+
     def juju_log(self, level, message):
-        self._run('juju-log', '--log-level', level, "--", message)
+        """Pass a log message on to the juju logger."""
+        for line in self.log_split(message):
+            self._run('juju-log', '--log-level', level, "--", line)
 
     def network_get(self, binding_name, relation_id=None):
         """Return network info provided by network-get for a given binding.
@@ -1533,6 +1803,19 @@ class _ModelBackend:
         """Create a pebble.Client instance from given socket path."""
         return pebble.Client(socket_path=socket_path)
 
+    def planned_units(self) -> int:
+        """Count of "planned" units that will run this application.
+
+        Includes the current unit in the count.
+
+        """
+        # The goal-state tool will return the information that we need. Goal state as a general
+        # concept is being deprecated, however, in favor of approaches such as the one that we use
+        # here.
+        app_state = self._run('goal-state', return_output=True, use_json=True)
+        # Planned units can be zero. We don't need to do error checking here.
+        return len(app_state.get('units', []))
+
 
 class _ModelBackendValidator:
     """Provides facilities for validating inputs and formatting them for model backends."""
@@ -1555,16 +1838,14 @@ class _ModelBackendValidator:
 
     @classmethod
     def format_metric_value(cls, value):
-        try:
-            decimal_value = decimal.Decimal.from_float(value)
-        except TypeError as e:
-            e2 = ModelError('invalid metric value {!r} provided:'
-                            ' must be a positive finite float'.format(value))
-            raise e2 from e
-        if decimal_value.is_nan() or decimal_value.is_infinite() or decimal_value < 0:
+        if not isinstance(value, (int, float)):
             raise ModelError('invalid metric value {!r} provided:'
                              ' must be a positive finite float'.format(value))
-        return str(decimal_value)
+
+        if math.isnan(value) or math.isinf(value) or value < 0:
+            raise ModelError('invalid metric value {!r} provided:'
+                             ' must be a positive finite float'.format(value))
+        return str(value)
 
     @classmethod
     def validate_label_value(cls, label, value):
