@@ -1,4 +1,4 @@
-# Copyright 2019-2020 Canonical Ltd.
+# Copyright 2019-2021 Canonical Ltd.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -13,9 +13,9 @@
 # limitations under the License.
 
 import abc
+import importlib.util
 import io
 import logging
-import logassert
 import os
 import platform
 import shutil
@@ -23,36 +23,37 @@ import subprocess
 import sys
 import tempfile
 import unittest
-import importlib.util
 import warnings
-import yaml
 from pathlib import Path
 from unittest.mock import patch
 
+import logassert
+import yaml
+
 from ops.charm import (
+    ActionEvent,
     CharmBase,
     CharmEvents,
     CharmMeta,
+    CollectMetricsEvent,
+    ConfigChangedEvent,
     HookEvent,
     InstallEvent,
-    StartEvent,
-    ConfigChangedEvent,
-    UpgradeCharmEvent,
-    UpdateStatusEvent,
     LeaderSettingsChangedEvent,
-    RelationJoinedEvent,
+    PebbleReadyEvent,
+    RelationBrokenEvent,
     RelationChangedEvent,
     RelationDepartedEvent,
-    RelationBrokenEvent,
     RelationEvent,
+    RelationJoinedEvent,
+    StartEvent,
     StorageAttachedEvent,
-    ActionEvent,
-    CollectMetricsEvent,
+    UpdateStatusEvent,
+    UpgradeCharmEvent,
     WorkloadEvent,
-    PebbleReadyEvent,
 )
 from ops.framework import Framework, StoredStateData
-from ops.main import main, CHARM_STATE_FILE, _should_use_controller_storage
+from ops.main import CHARM_STATE_FILE, _should_use_controller_storage, main
 from ops.storage import SQLiteStorage
 from ops.version import version
 
@@ -709,6 +710,46 @@ class TestMainWithNoDispatch(_TestMain, unittest.TestCase):
         event_file = self.JUJU_CHARM_DIR / rel_path
         # Note that sys.executable is used to make sure we are using the same
         # interpreter for the child process to support virtual environments.
+        fake_script(
+            self,
+            "storage-get",
+            """
+            if [ "$1" = "-s" ]; then
+                id=${2#*/}
+                key=${2%/*}
+                echo "\\"/var/srv/${key}/${id}\\"" # NOQA: test_quote_backslashes
+            elif [ "$1" = '--help' ]; then
+                printf '%s\\n' \\
+                'Usage: storage-get [options] [<key>]' \\
+                '   ' \\
+                'Summary:' \\
+                'print information for storage instance with specified id' \\
+                '   ' \\
+                'Options:' \\
+                '--format  (= smart)' \\
+                '    Specify output format (json|smart|yaml)' \\
+                '-o, --output (= "")' \\
+                '    Specify an output file' \\
+                '-s  (= test-stor/0)' \\
+                '    specify a storage instance by id' \\
+                '   ' \\
+                'Details:' \\
+                'When no <key> is supplied, all keys values are printed.'
+            else
+                # Return the same path for all disks since `storage-get`
+                # on attach and detach takes no parameters and is not
+                # deterministically faked with fake_script
+                exit 1
+            fi
+            """,
+        )
+        fake_script(
+            self,
+            "storage-list",
+            """
+            echo '["disks/0"]'
+            """,
+        )
         subprocess.run(
             [sys.executable, str(event_file)],
             check=True, env=env, cwd=str(self.JUJU_CHARM_DIR))
@@ -739,7 +780,7 @@ class TestMainWithNoDispatch(_TestMain, unittest.TestCase):
                     self.assertFalse(hook_path.is_symlink())
                 else:
                     # hooks are not symlinks, and this hook is not one of the initial ones
-                    # check that it's a symlink to the inital ones
+                    # check that it's a symlink to the initial ones
                     self.assertTrue(hook_path.is_symlink())
                     self.assertEqual(os.readlink(str(hook_path)), event_spec.event_name)
 
@@ -785,7 +826,9 @@ class _TestMainWithDispatch(_TestMain):
     has_dispatch = True
 
     def test_setup_event_links(self):
-        """Test auto-creation of symlinks caused by initial events does _not_ happen when using dispatch.
+        """Test auto-creation of symlinks.
+
+        Symlink creation caused by initial events should _not_ happen when using dispatch.
         """
         all_event_hooks = ['hooks/' + e.replace("_", "-")
                            for e in self.charm_module.Charm.on.events().keys()]
@@ -957,6 +1000,46 @@ class TestMainWithDispatch(_TestMainWithDispatch, unittest.TestCase):
         env['JUJU_DISPATCH_PATH'] = str(rel_path)
         env['JUJU_VERSION'] = '2.8.0'
         dispatch = self.JUJU_CHARM_DIR / 'dispatch'
+        fake_script(
+            self,
+            "storage-get",
+            """
+            if [ "$1" = "-s" ]; then
+                id=${2#*/}
+                key=${2%/*}
+                echo "\\"/var/srv/${key}/${id}\\"" # NOQA: test_quote_backslashes
+            elif [ "$1" = '--help' ]; then
+                printf '%s\\n' \\
+                'Usage: storage-get [options] [<key>]' \\
+                '   ' \\
+                'Summary:' \\
+                'print information for storage instance with specified id' \\
+                '   ' \\
+                'Options:' \\
+                '--format  (= smart)' \\
+                '    Specify output format (json|smart|yaml)' \\
+                '-o, --output (= "")' \\
+                '    Specify an output file' \\
+                '-s  (= test-stor/0)' \\
+                '    specify a storage instance by id' \\
+                '   ' \\
+                'Details:' \\
+                'When no <key> is supplied, all keys values are printed.'
+            else
+                # Return the same path for all disks since `storage-get`
+                # on attach and detach takes no parameters and is not
+                # deterministically faked with fake_script
+                exit 1
+            fi
+            """,
+        )
+        fake_script(
+            self,
+            "storage-list",
+            """
+            echo '["disks/0"]'
+            """,
+        )
         subprocess.run(
             [sys.executable, str(dispatch)],
             # stdout=self.stdout,
@@ -987,6 +1070,46 @@ class TestMainWithDispatchAsScript(_TestMainWithDispatch, unittest.TestCase):
     def _call_event(self, rel_path, env):
         env['JUJU_DISPATCH_PATH'] = str(rel_path)
         env['JUJU_VERSION'] = '2.8.0'
+        fake_script(
+            self,
+            "storage-get",
+            """
+            if [ "$1" = "-s" ]; then
+                id=${2#*/}
+                key=${2%/*}
+                echo "\\"/var/srv/${key}/${id}\\"" # NOQA: test_quote_backslashes
+            elif [ "$1" = '--help' ]; then
+                printf '%s\\n' \\
+                'Usage: storage-get [options] [<key>]' \\
+                '   ' \\
+                'Summary:' \\
+                'print information for storage instance with specified id' \\
+                '   ' \\
+                'Options:' \\
+                '--format  (= smart)' \\
+                '    Specify output format (json|smart|yaml)' \\
+                '-o, --output (= "")' \\
+                '    Specify an output file' \\
+                '-s  (= test-stor/0)' \\
+                '    specify a storage instance by id' \\
+                '   ' \\
+                'Details:' \\
+                'When no <key> is supplied, all keys values are printed.'
+            else
+                # Return the same path for all disks since `storage-get`
+                # on attach and detach takes no parameters and is not
+                # deterministically faked with fake_script
+                exit 1
+            fi
+            """,
+        )
+        fake_script(
+            self,
+            "storage-list",
+            """
+            echo '["disks/0"]'
+            """,
+        )
         dispatch = (self.JUJU_CHARM_DIR / 'dispatch').with_suffix(self.suffix)
         subprocess.check_call([str(dispatch)], env=env, cwd=str(self.JUJU_CHARM_DIR))
 
