@@ -13,23 +13,21 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections import OrderedDict
-import json
 import ipaddress
+import json
 import os
 import pathlib
-from textwrap import dedent
 import unittest
-
-import ops.model
-import ops.charm
-import ops.pebble
-import ops.testing
-from ops.charm import RelationMeta, RelationRole
-
-from ops._private import yaml
-
+from collections import OrderedDict
 from test.test_helpers import fake_script, fake_script_calls
+from textwrap import dedent
+
+import ops.charm
+import ops.model
+import ops.testing
+from ops._private import yaml
+from ops.charm import RelationMeta, RelationRole
+from ops.pebble import APIError, ServiceInfo
 
 
 class TestModel(unittest.TestCase):
@@ -49,6 +47,14 @@ class TestModel(unittest.TestCase):
             resources:
               foo: {type: file, filename: foo.txt}
               bar: {type: file, filename: bar.txt}
+        ''', config='''
+        options:
+            foo:
+                type: string
+            bar:
+                type: int
+            qux:
+                type: bool
         ''')
         self.addCleanup(self.harness.cleanup)
         self.relation_id_db0 = self.harness.add_relation('db0', 'db')
@@ -484,7 +490,7 @@ class TestModel(unittest.TestCase):
         self.harness.add_resource('foo', 'foo contents\n')
         self.harness.add_resource('bar', '')
 
-        with self.assertRaises(RuntimeError):
+        with self.assertRaises(NameError):
             self.harness.model.resources.fetch('qux')
 
         self.assertEqual(self.harness.model.resources.fetch('foo').name, 'foo.txt')
@@ -742,13 +748,102 @@ class TestModel(unittest.TestCase):
         with self.assertRaises(AttributeError):
             self.model.storages = {}
 
-    def resetBackendCalls(self):
+    def resetBackendCalls(self):  # noqa: N802
         self.harness._get_backend_calls(reset=True)
 
-    def assertBackendCalls(self, expected, *, reset=True):
+    def assertBackendCalls(self, expected, *, reset=True):  # noqa: N802
         self.assertEqual(expected, self.harness._get_backend_calls(reset=reset))
 
 
+class TestApplication(unittest.TestCase):
+
+    def setUp(self):
+        self.harness = ops.testing.Harness(ops.charm.CharmBase, meta='''
+            name: myapp
+            provides:
+              db0:
+                interface: db0
+            requires:
+              db1:
+                interface: db1
+            peers:
+              db2:
+                interface: db2
+            resources:
+              foo: {type: file, filename: foo.txt}
+              bar: {type: file, filename: bar.txt}
+        ''')
+        self.peer_rel_id = self.harness.add_relation('db2', 'db2')
+        self.app = self.harness.model.app
+
+    def test_planned_units(self):
+        rel_id = self.peer_rel_id
+
+        # Test that we always count ourself.
+        self.assertEqual(self.app.planned_units(), 1)
+
+        # Add some units, and verify count.
+        self.harness.add_relation_unit(rel_id, 'myapp/1')
+        self.harness.add_relation_unit(rel_id, 'myapp/2')
+
+        self.assertEqual(self.app.planned_units(), 3)
+
+        self.harness.add_relation_unit(rel_id, 'myapp/3')
+        self.assertEqual(self.app.planned_units(), 4)
+
+        # And remove a unit
+        self.harness.remove_relation_unit(rel_id, 'myapp/2')
+
+        self.assertEqual(self.app.planned_units(), 3)
+
+    def test_planned_units_user_set(self):
+
+        self.harness.set_planned_units(1)
+        self.assertEqual(self.app.planned_units(), 1)
+
+        self.harness.set_planned_units(2)
+        self.assertEqual(self.app.planned_units(), 2)
+
+        self.harness.set_planned_units(100)
+        self.assertEqual(self.app.planned_units(), 100)
+
+    def test_planned_units_garbage_values(self):
+        # Planned units should be a positive integer, or zero.
+        with self.assertRaises(TypeError):
+            self.harness.set_planned_units(-1)
+        # Verify that we didn't set our value before raising the error.
+        self.assertTrue(self.harness._backend._planned_units is None)
+        # Verify that we still get the default value back from .planned_units.
+        self.assertEqual(self.app.planned_units(), 1)
+
+        with self.assertRaises(TypeError):
+            self.harness.set_planned_units("foo")
+
+        with self.assertRaises(TypeError):
+            self.harness.set_planned_units(-3423000102312321090)
+
+    def test_planned_units_override(self):
+        """Verify that we override the calculated value of planned_units when we set it manually.
+
+        When a charm author writes a test that explicitly calls set_planned_units, we assume that
+        their intent is to override the calculated return value. Often, this will be because the
+        charm author is composing a charm without peer relations, and the harness's count of
+        planned units, which is based on the number of peer relations, will not be accurate.
+        """
+        peer_id = self.peer_rel_id
+
+        self.harness.set_planned_units(10)
+        self.harness.add_relation_unit(peer_id, 'myapp/1')
+        self.harness.add_relation_unit(peer_id, 'myapp/2')
+        self.harness.add_relation_unit(peer_id, 'myapp/3')
+
+        self.assertEqual(self.app.planned_units(), 10)
+
+        # Verify that we can clear the override.
+        self.harness.reset_planned_units()
+        self.assertEqual(self.app.planned_units(), 4)  # self + 3 peers
+
+
 class TestContainers(unittest.TestCase):
     def setUp(self):
         meta = ops.charm.CharmMeta.from_yaml("""
@@ -816,6 +911,14 @@ containers:
         self.container.autostart()
         self.assertEqual(self.pebble.requests, [('autostart',)])
 
+    def test_replan(self):
+        self.container.replan()
+        self.assertEqual(self.pebble.requests, [('replan',)])
+
+    def test_get_system_info(self):
+        self.container.can_connect()
+        self.assertEqual(self.pebble.requests, [('get_system_info',)])
+
     def test_start(self):
         self.container.start('foo')
         self.container.start('foo', 'bar')
@@ -824,6 +927,10 @@ containers:
             ('start', ('foo', 'bar')),
         ])
 
+    def test_start_no_arguments(self):
+        with self.assertRaises(TypeError):
+            self.container.start()
+
     def test_stop(self):
         self.container.stop('foo')
         self.container.stop('foo', 'bar')
@@ -832,6 +939,54 @@ containers:
             ('stop', ('foo', 'bar')),
         ])
 
+    def test_stop_no_arguments(self):
+        with self.assertRaises(TypeError):
+            self.container.stop()
+
+    def test_restart(self):
+        self.container.restart('foo')
+        self.container.restart('foo', 'bar')
+        self.assertEqual(self.pebble.requests, [
+            ('restart', ('foo',)),
+            ('restart', ('foo', 'bar')),
+        ])
+
+    def test_restart_fallback(self):
+        def restart_services(services):
+            self.pebble.requests.append(('restart', services))
+            raise APIError({}, 400, "", "")
+
+        self.pebble.restart_services = restart_services
+        # Setup the Pebble client to respond to a call to get_services()
+        self.pebble.responses.append([
+            ServiceInfo.from_dict({'name': 'foo', 'startup': 'enabled', 'current': 'active'}),
+            ServiceInfo.from_dict({'name': 'bar', 'startup': 'enabled', 'current': 'inactive'}),
+        ])
+
+        self.container.restart('foo', 'bar')
+        self.assertEqual(self.pebble.requests, [
+            # This is the first request, which in real life fails with APIError on older versions
+            ('restart', ('foo', 'bar')),
+            # Next the code should loop over the started services, and stop them
+            ('get_services', ('foo', 'bar')),
+            ('stop', ('foo',)),
+            # Then start all the specified services
+            ('start', ('foo', 'bar'))
+        ])
+
+    def test_restart_fallback_non_400_error(self):
+        def restart_services(services):
+            raise APIError({}, 500, "", "")
+
+        self.pebble.restart_services = restart_services
+        with self.assertRaises(ops.pebble.APIError) as cm:
+            self.container.restart('foo')
+        self.assertEqual(cm.exception.code, 500)
+
+    def test_restart_no_arguments(self):
+        with self.assertRaises(TypeError):
+            self.container.restart()
+
     def test_type_errors(self):
         meta = ops.charm.CharmMeta.from_yaml("""
 name: k8s-charm
@@ -1005,6 +1160,61 @@ containers:
             ('remove_path', '/path/2', True),
         ])
 
+    def test_bare_can_connect_call(self):
+        self.pebble.responses.append('dummy')
+        self.assertTrue(self.container.can_connect())
+
+    def test_exec(self):
+        self.pebble.responses.append('fake_exec_process')
+        p = self.container.exec(
+            ['echo', 'foo'],
+            environment={'K1': 'V1', 'K2': 'V2'},
+            working_dir='WD',
+            timeout=10.5,
+            user_id=1000,
+            user='bob',
+            group_id=1000,
+            group='staff',
+            stdin='STDIN',
+            stdout='STDOUT',
+            stderr='STDERR',
+            encoding=None,
+            combine_stderr=True,
+        )
+        self.assertEqual(self.pebble.requests, [
+            ('exec', ['echo', 'foo'], dict(
+                environment={'K1': 'V1', 'K2': 'V2'},
+                working_dir='WD',
+                timeout=10.5,
+                user_id=1000,
+                user='bob',
+                group_id=1000,
+                group='staff',
+                stdin='STDIN',
+                stdout='STDOUT',
+                stderr='STDERR',
+                encoding=None,
+                combine_stderr=True,
+            ))
+        ])
+        self.assertEqual(p, 'fake_exec_process')
+
+    def test_send_signal(self):
+        with self.assertRaises(TypeError):
+            self.container.send_signal('SIGHUP')
+
+        self.container.send_signal('SIGHUP', 's1')
+        self.assertEqual(self.pebble.requests, [
+            ('send_signal', 'SIGHUP', ('s1',)),
+        ])
+        self.pebble.requests = []
+
+        self.container.send_signal('SIGHUP', 's1', 's2')
+        self.assertEqual(self.pebble.requests, [
+            ('send_signal', 'SIGHUP', ('s1', 's2')),
+        ])
+        self.pebble.requests = []
+
 
 class MockPebbleBackend(ops.model._ModelBackend):
     def get_pebble(self, socket_path):
@@ -1020,12 +1230,21 @@ class MockPebbleClient:
     def autostart_services(self):
         self.requests.append(('autostart',))
 
+    def get_system_info(self):
+        self.requests.append(('get_system_info',))
+
+    def replan_services(self):
+        self.requests.append(('replan',))
+
     def start_services(self, service_names):
         self.requests.append(('start', service_names))
 
     def stop_services(self, service_names):
         self.requests.append(('stop', service_names))
 
+    def restart_services(self, service_names):
+        self.requests.append(('restart', service_names))
+
     def add_layer(self, label, layer, combine=False):
         if isinstance(layer, dict):
             layer = ops.pebble.Layer(layer).to_yaml()
@@ -1062,6 +1281,13 @@ class MockPebbleClient:
     def remove_path(self, path, *, recursive=False):
         self.requests.append(('remove_path', path, recursive))
 
+    def exec(self, command, **kwargs):
+        self.requests.append(('exec', command, kwargs))
+        return self.responses.pop(0)
+
+    def send_signal(self, signal, service_names):
+        self.requests.append(('send_signal', signal, service_names))
+
 
 class TestModelBindings(unittest.TestCase):
 
@@ -1623,8 +1849,8 @@ class TestModelBackend(unittest.TestCase):
         fake_script(self, 'action-set', 'echo fooerror >&2 ; exit 1')
         with self.assertRaises(ops.model.ModelError):
             self.backend.action_set(OrderedDict([('foo', 'bar'), ('dead', 'beef cafe')]))
-        calls = [["action-set", "foo=bar", "dead=beef cafe"]]
-        self.assertEqual(fake_script_calls(self, clear=True), calls)
+        self.assertCountEqual(
+            ["action-set", "dead=beef cafe", "foo=bar"], fake_script_calls(self, clear=True)[0])
 
     def test_action_log_error(self):
         fake_script(self, 'action-get', '')
@@ -1644,8 +1870,45 @@ class TestModelBackend(unittest.TestCase):
     def test_action_set(self):
         fake_script(self, 'action-get', 'exit 1')
         fake_script(self, 'action-set', 'exit 0')
-        self.backend.action_set(OrderedDict([('x', 'dead beef'), ('y', 1)]))
-        self.assertEqual(fake_script_calls(self), [['action-set', 'x=dead beef', 'y=1']])
+        self.backend.action_set({'x': 'dead beef', 'y': 1})
+        self.assertCountEqual(['action-set', 'x=dead beef', 'y=1'], fake_script_calls(self)[0])
+
+    def test_action_set_key_validation(self):
+        with self.assertRaises(ValueError):
+            self.backend.action_set({'X': 'dead beef', 'y': 1})
+        with self.assertRaises(ValueError):
+            self.backend.action_set({'some&key': 'dead beef', 'y': 1})
+        with self.assertRaises(ValueError):
+            self.backend.action_set({'someKey': 'dead beef', 'y': 1})
+        with self.assertRaises(ValueError):
+            self.backend.action_set({'some_key': 'dead beef', 'y': 1})
+
+    def test_action_set_nested(self):
+        fake_script(self, 'action-get', 'exit 1')
+        fake_script(self, 'action-set', 'exit 0')
+        self.backend.action_set({'a': {'b': 1, 'c': 2}, 'd': 3})
+        self.assertCountEqual(['action-set', 'a.b=1', 'a.c=2', 'd=3'], fake_script_calls(self)[0])
+
+    def test_action_set_more_nested(self):
+        fake_script(self, 'action-get', 'exit 1')
+        fake_script(self, 'action-set', 'exit 0')
+        self.backend.action_set({'a': {'b': 1, 'c': 2, 'd': {'e': 3}}, 'f': 4})
+        self.assertCountEqual(
+            ['action-set', 'a.b=1', 'a.c=2', 'a.d.e=3', 'f=4'], fake_script_calls(self)[0])
+
+    def test_action_set_dotted_dict(self):
+        fake_script(self, 'action-get', 'exit 1')
+        fake_script(self, 'action-set', 'exit 0')
+        self.backend.action_set({'a.b': 1, 'a': {'c': 2}, 'd': 3})
+        self.assertCountEqual(['action-set', 'a.b=1', 'a.c=2', 'd=3'], fake_script_calls(self)[0])
+
+    def test_action_set_duplicated_keys(self):
+        fake_script(self, 'action-get', 'exit 1')
+        fake_script(self, 'action-set', 'exit 0')
+        with self.assertRaises(ValueError):
+            self.backend.action_set({'a.b': 1, 'a': {'b': 2}, 'd': 3})
+        with self.assertRaises(ValueError):
+            self.backend.action_set({'a': {'b': 1, 'c': 2, 'd': {'e': 3}}, 'f': 4, 'a.d.e': 'foo'})
 
     def test_action_fail(self):
         fake_script(self, 'action-get', 'exit 1')
