@@ -19,31 +19,23 @@ import os
 import pathlib
 import sys
 import tempfile
+from test.test_helpers import BaseTestCase, fake_script, fake_script_calls
 from textwrap import dedent
 
 import yaml
 
-from ops import (
-    framework,
-    storage,
-)
-from test.test_helpers import (
-    BaseTestCase,
-    fake_script,
-    fake_script_calls,
-)
+from ops import framework, storage
 
 
 class StoragePermutations(abc.ABC):
 
     def create_framework(self) -> framework.Framework:
-        """Create a Framework that we can use to test the backend storage.
-        """
+        """Create a Framework that we can use to test the backend storage."""
         return framework.Framework(self.create_storage(), None, None, None)
 
     @abc.abstractmethod
     def create_storage(self) -> storage.SQLiteStorage:
-        """Create a Storage backend that we can interact with"""
+        """Create a Storage backend that we can interact with."""
         return NotImplemented
 
     def test_save_and_load_snapshot(self):
@@ -215,7 +207,7 @@ class TestSQLiteStorage(StoragePermutations, BaseTestCase):
 
 
 def setup_juju_backend(test_case, state_file):
-    """Create fake scripts for pretending to be state-set and state-get"""
+    """Create fake scripts for pretending to be state-set and state-get."""
     template_args = {
         'executable': str(pathlib.Path(sys.executable).as_posix()),
         'pthpth': repr(os.path.dirname(pathlib.__file__))[1:-1],
@@ -313,7 +305,7 @@ class TestSimpleLoader(BaseTestCase):
         parsed = yaml.load(raw, Loader=storage._SimpleLoader)
         self.assertEqual(parsed, (1, 'tuple'))
 
-    def assertRefused(self, obj):
+    def assertRefused(self, obj):  # noqa: N802
         # We shouldn't allow them to be written
         with self.assertRaises(yaml.representer.RepresenterError):
             yaml.dump(obj, Dumper=storage._SimpleDumper)
